/******************************************************************************
*
* Nagios check_mysql_perf plugin
*
* License: GPL
* Copyright (c) 2008 ConSol GmbH, Gerhard Lausser
*
* Last Modified: $Date: 2007-01-28 21:46:41 +0000 (Sun, 28 Jan 2007) $
*
* Description:
*
* This file contains the check_mysql_perf plugin
*
*  This plugin is for running arbitrary SQL and checking the results
*
* License Information:
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
* CHECK_MYSQL_PERF.C
*
*
******************************************************************************/

const char *progname = "check_mysql_perf";
const char *revision = "1.2";
const char *copyright = "2008";
const char *email = "gerhard.lausser@consol.de";

#include "common.h"
#include "utils.h"
#include "utils_base.h"
#include "netutils.h"
#include "my_utils.h"

#include <mysql.h>
#include <errmsg.h>

#define MAXPLUGINOUTPUTLENGTH 255
#define MAXPERFDATALENGTH 255
#define pluginexit(code, ...) { snprintf(exitmessagestr, MAXPLUGINOUTPUTLENGTH, __VA_ARGS__); exitcode = code; }
#define perfdata(...) snprintf(perfdatastr, MAXPLUGINOUTPUTLENGTH, __VA_ARGS__)
#define prdebug(...) if (verbose) fprintf(stderr, __VA_ARGS__)


char *db_user = NULL;
char *db_host = NULL;
char *db_pass = NULL;
char *db = NULL;
unsigned int db_port = MYSQL_PORT;

int process_arguments (int, char **);
int validate_arguments (void);
void print_help (void);
void print_usage (void);

void save_variable (char *, char *);
char *load_variable (char *, long *);
void force_thresholds (thresholds **, char **, char **, const char *, const char *);
char *run_query (MYSQL *, char *, int);
char *mysql_status (MYSQL *, char *);
char *mysql_variable (MYSQL *, char *);
char *mysql_status_historical (MYSQL *, char *, long *, int);


char *sql_query = NULL;
int verbose = 0;
int debug = 0;
thresholds *my_thresholds = NULL;
char *userwarning = NULL;
char *usercritical = NULL;

char *runident = NULL;
char *mode = NULL;
int interval = -1;
int absolute = 0;


int
main (int argc, char **argv)
{

	MYSQL mysql;
	
	double value;
	char *error = NULL;
	char *result = NULL;
	int status = STATE_UNKNOWN;
	struct timeval timer_start, timer_stop;
	char exitmessagestr[MAXPLUGINOUTPUTLENGTH];
	char perfdatastr[MAXPERFDATALENGTH] = "";
	char *severities[4] = {"OK", "WARNING", "CRITICAL", "UNKNOWN"};
	int exitcode = STATE_UNKNOWN;

	if (process_arguments (argc, argv) == ERROR)
		usage4 (_("Could not parse arguments"));

	/* build a unique string identifying this plugin run */
	asprintf(&runident, "%s_%s_%s_%s_%d",
	    progname,
	    db_host == NULL ? "" : db_host,
	    db_user == NULL ? "" : db_user,
	    db == NULL ? "" : db,
	    db_port);

	/* initialize mysql  */
	mysql_init (&mysql);

	mysql_options(&mysql,MYSQL_READ_DEFAULT_GROUP,"client");

	gettimeofday(&timer_start, 0);
	/* establish a connection to the server and error checking */
	if (!mysql_real_connect(&mysql,db_host,db_user,db_pass,db,db_port,NULL,0)) {
		pluginexit(STATE_CRITICAL, "Cannot connect to database: Error: %s", mysql_error(&mysql));
	} else {
		if (strcmp(mode, "slave-lag") == 0) {
			int lag = 0;
			force_thresholds(&my_thresholds, &userwarning, &usercritical, "10", "20");
			result = run_query(&mysql, "show slave status", 32);
			if (result == NULL) {
				pluginexit(STATE_CRITICAL, "Slave lag NULL");
			} else {
				lag = atoi(result);
				free(result);
				status = get_status((double)lag, my_thresholds);
				if (status == STATE_OK) {
					pluginexit(status, "Slave lag %d is inside tolerance.", lag);
				} else if (status == STATE_WARNING) {
					pluginexit(status, "Slave lag %d sec exceeds warning threshold %s sec",
					    lag, userwarning);
				} else if (status == STATE_CRITICAL) {
					pluginexit(status, "Slave lag %d sec exceeds critical threshold %s sec",
					    lag, usercritical);
				}
			}
			perfdata("slave_lag=%d;%s;%s", lag, userwarning, usercritical);
		} else if (strcmp(mode, "slave-io-running") == 0) {
			result = run_query(&mysql, "show slave status", 10);
			if (result == NULL) {
				pluginexit(STATE_CRITICAL, "Slave io not running");
			} else if (strcmp(result, "Yes") != 0) {
				pluginexit(STATE_CRITICAL, "Slave io not running: result is %s", result);
			} else {
				pluginexit(STATE_OK, "Slave io is running.");
			}
			perfdata("slave_io_running=%d", ((result == NULL) ? 0 : ((strcmp(result, "Yes") == 0) ? 1 : 0)));
		} else if (strcmp(mode, "slave-sql-running") == 0) {
			result = run_query(&mysql, "show slave status", 11);
			if (result == NULL) {
				pluginexit(STATE_CRITICAL, "Slave sql not running");
			} else if (strcmp(result, "Yes") != 0) {
				pluginexit(STATE_CRITICAL, "Slave sql not running: result is %s", result);
			} else {
				pluginexit(STATE_OK, "Slave sql is running.");
			}
			perfdata("slave_sql_running=%d", ((result == NULL) ? 0 : ((strcmp(result, "Yes") == 0) ? 1 : 0)));
		} else if (strcmp(mode, "threads-connected") == 0) {
			int  conn = 0;
			char *result_max = NULL;
			int max = 0;
			force_thresholds(&my_thresholds, &userwarning, &usercritical, "10", "20");
			result = mysql_status(&mysql, "Threads_connected");
			/* use this for a future threads-connected-pct
			result_max = run_query(&mysql, "show  global variables like 'Max_connections'", 1);
			*/
			if (result == NULL) {
				pluginexit(STATE_CRITICAL, "No connection information available.");
			} else {
				conn = atoi(result);
				free(result);
				status = get_status((double)conn, my_thresholds);
				if (status == STATE_OK) {
					pluginexit(status, "Threads_connected: %d", conn);
				} else if (status == STATE_WARNING) {
					pluginexit(status, "Threads_connected %d exceeds warning threshold %s",
					    conn, userwarning);
				} else if (status == STATE_CRITICAL) {
					pluginexit(status, "Threads_connected %d exceeds critical threshold %s",
					    conn, usercritical);
				}
			}
			perfdata("threads_connected=%d;%s;%s", conn, userwarning, usercritical);
                } else if (strcmp(mode, "threadcache-hitrate") == 0) {
                        char *q, *c;
                        long threads_created = 0.0;
                        long connections = 0.0;
                        double threadcache_hitrate = 0.0;

                        force_thresholds(&my_thresholds, &userwarning, &usercritical, "90:", "80:");
                        c = mysql_status(&mysql, "Threads_created");
                        q = mysql_status(&mysql, "Connections");
                        if ((c == NULL) || (q == NULL)) {
                                pluginexit(STATE_CRITICAL, "No connection information available.");
                        } else {
                                threads_created = atol(c);
                                connections = atol(q);
                                free(c);
                                free(q);
                                if ((connections) > 0) {
                                        threadcache_hitrate = 
                                            100 - (threads_created * 100.0 / connections);
                                }
                                status = get_status(threadcache_hitrate, my_thresholds);
                                pluginexit(status, "Thread Cache Hitrate at %.2f%%", threadcache_hitrate);
                        }
                        perfdata("threadcache_hitrate=%.2f%%;%s;%s", threadcache_hitrate, userwarning, usercritical);
		} else if ((strcmp(mode, "querycache-hitrate") == 0) ||
		    (strcmp(mode, "qcache-hitrate") == 0)) {
			char *q, *c;
			long qcache_hits = 0.0;
			long com_select = 0.0;
			double qcache_hitrate = 0.0;

			force_thresholds(&my_thresholds, &userwarning, &usercritical, "90:", "80:");
			c = mysql_status(&mysql, "Com_select");
      			q = mysql_status(&mysql, "Qcache_hits");
			if ((c == NULL) || (q == NULL)) {
				pluginexit(STATE_CRITICAL, "No connection information available.");
			} else {
				com_select = atol(c);
				qcache_hits = atol(q);
				free(c);
				free(q);
				if ((qcache_hits + com_select) > 0) {
					qcache_hitrate =
					    (qcache_hits * 100.0) / (qcache_hits + com_select + 1.0);
				}
				status = get_status(qcache_hitrate, my_thresholds);
				pluginexit(status, "Query Cache Hitrate at %.2f%%", qcache_hitrate);
			}
			perfdata("qcache_hitrate=%.2f%%;%s;%s", qcache_hitrate, userwarning, usercritical);
		} else if (strcmp(mode, "keycache-hitrate") == 0) {
			char *q, *r;
			double keycache_hitrate = 0.0;
			long  key_reads = 0;
			long  key_read_requests = 0;

			force_thresholds(&my_thresholds, &userwarning, &usercritical, "99:", "95:");
			r = mysql_status(&mysql, "Key_reads");
			q = mysql_status(&mysql, "Key_read_requests");
			if ((r == NULL) || (q == NULL)) {
				pluginexit(STATE_CRITICAL, "No connection information available.");
			} else {
				key_reads = atol(r);
				key_read_requests = atol(q);
				free(r);
				free(q);
				if (key_read_requests > 0) {
					keycache_hitrate = 100 - (key_reads * 100.0 / key_read_requests);
				}
				status = get_status(keycache_hitrate, my_thresholds);
				pluginexit(status, "MyISAM Key Cache Hitrate at %.2f%%", keycache_hitrate);
			}
			perfdata("keycache_hitrate=%.2f%%;%s;%s", keycache_hitrate, userwarning, usercritical);
                } else if (strcmp(mode, "bufferpool-hitrate") == 0) {
                        char *q, *r;
                        double bufferpool_hitrate = 0.0;
                        long  innodb_buffer_pool_reads = 0;
                        long  innodb_buffer_pool_read_requests = 0;

                        force_thresholds(&my_thresholds, &userwarning, &usercritical, "99:", "95:");
                        r = mysql_status(&mysql, "Innodb_buffer_pool_reads");
                        q = mysql_status(&mysql, "Innodb_buffer_pool_read_requests");
                        if ((r == NULL) || (q == NULL)) {
                                pluginexit(STATE_CRITICAL, "No connection information available.");
                        } else {
                                innodb_buffer_pool_reads = atol(r);
                                innodb_buffer_pool_read_requests = atol(q);
                                free(r);
                                free(q);
                                if (innodb_buffer_pool_read_requests > 0) {
                                        bufferpool_hitrate = 100 - (innodb_buffer_pool_reads * 100.0 / innodb_buffer_pool_read_requests);
                                }
                                status = get_status(bufferpool_hitrate, my_thresholds);
                                pluginexit(status, "Innodb Buffer Pool Hitrate at %.2f%%", bufferpool_hitrate);
                        }
                        perfdata("bufferpool_hitrate=%.2f%%;%s;%s", bufferpool_hitrate, userwarning, usercritical);
                } else if (strcmp(mode, "tablecache-hitrate") == 0) {
                        char *q, *r;
                        double tablecache_hitrate = 100.0;
                        long  open_tables = 0;
                        long  opened_tables = 0;

                        force_thresholds(&my_thresholds, &userwarning, &usercritical, "99:", "95:");
                        r = mysql_status(&mysql, "Open_tables");
                        q = mysql_status(&mysql, "Opened_tables");
                        if ((r == NULL) || (q == NULL)) {
                                pluginexit(STATE_CRITICAL, "No connection information available.");
                        } else {
                                open_tables = atol(r);
                                opened_tables = atol(q);
                                free(r);
                                free(q);
                                if (opened_tables > 0) {
                                        tablecache_hitrate = open_tables * 100 / opened_tables;
				}
                                status = get_status(tablecache_hitrate, my_thresholds);
                                pluginexit(status, "Table Cache Hitrate at %.2f%%", tablecache_hitrate);
                        }
                        perfdata("tablecache_hitrate=%.2f%%;%s;%s", tablecache_hitrate, userwarning, usercritical);

                } else if (strcmp(mode, "table-lock-contention") == 0) {
                        char *q, *r;
                        double table_lock_contention = 0.0;
                        long  table_locks_waited = 0;
                        long  table_locks_immediate = 0;

                        force_thresholds(&my_thresholds, &userwarning, &usercritical, "1", "2");
                        r = mysql_status(&mysql, "Table_locks_waited");
                        q = mysql_status(&mysql, "Table_locks_immediate");
                        if ((r == NULL) || (q == NULL)) {
                                pluginexit(STATE_CRITICAL, "No connection information available.");
                        } else {
                                table_locks_waited = atol(r);
                                table_locks_immediate = atol(q);
                                free(r);
                                free(q);
                                if ((table_locks_waited + table_locks_immediate) > 0) {
                                        table_lock_contention = table_locks_waited * 100 / (table_locks_waited + table_locks_immediate);
                                }
                                status = get_status(table_lock_contention, my_thresholds);
                                pluginexit(status, "Table lock Contention at %.2f%%", table_lock_contention);
                        }
                        perfdata("table_lock_contention=%.2f%%;%s;%s", table_lock_contention, userwarning, usercritical);
		} else if (strcmp(mode, "connection-time") == 0) {
			double elapsed = 0.0;
			gettimeofday(&timer_stop, 0);
			force_thresholds(&my_thresholds, &userwarning, &usercritical, "2", "10");
			elapsed = ((timer_stop.tv_sec * 1000000.0 + timer_stop.tv_usec) -
			    (timer_start.tv_sec * 1000000.0 + timer_start.tv_usec)) / 1000000.0;
			status = get_status(elapsed, my_thresholds);
			pluginexit(status, "Connection Time %.3f seconds", elapsed);
			perfdata("connection_time=%.2f;%s;%s", elapsed, userwarning, usercritical);
		} else if ((strcmp(mode, "temp-disk-tables") == 0) ||
		    (strcmp(mode, "tmp-disk-tables") == 0)) {
			char *q, *r;
			double pct_tmp_on_disk = 0.0;
			long  created_tmp_tables = 0;
			long  created_tmp_disk_tables = 0;

			force_thresholds(&my_thresholds, &userwarning, &usercritical, "25", "50");
			r = mysql_status(&mysql, "Created_tmp_tables");
			q = mysql_status(&mysql, "Created_tmp_disk_tables");
			if ((r == NULL) || (q == NULL)) {
				pluginexit(STATE_CRITICAL, "No connection information available.");
			} else {
				created_tmp_tables = atol(r);
				created_tmp_disk_tables = atol(q);
				free(r);
				free(q);
				if (created_tmp_tables > 0) {
					pct_tmp_on_disk = created_tmp_disk_tables * 100.0 / created_tmp_tables;
				}
				status = get_status(pct_tmp_on_disk, my_thresholds);
				pluginexit(status, "%.2f%% of %ld temp tables were created on disk",
				    pct_tmp_on_disk, created_tmp_tables);
			}
			perfdata("pct_tmp_table_on_disk=%.2f%%;%s;%s", pct_tmp_on_disk, userwarning, usercritical);
		} else if (strcmp(mode, "slow-queries") == 0) {
			char *r, *q;
			double slow_queries_per_sec = 0.0;
			long  slow_queries_delta = 0;
			long  slow_queries = 0;
			long  last_slow_queries = 0;
			time_t secsago;
			
			interval = (interval > 0) ? interval : 0;
			force_thresholds(&my_thresholds, &userwarning, &usercritical, "1", "10");
			q = mysql_status_historical(&mysql, "Slow_queries", &secsago, interval);
			r = mysql_status(&mysql, "Slow_queries");
			if (r == NULL) {
				pluginexit(STATE_CRITICAL, "No connection information available.");
			} else if (q == NULL) {
				slow_queries = atol(r);
				free(r);
				pluginexit(STATE_UNKNOWN, "no historical data");
			} else {
				slow_queries = atol(r);
				last_slow_queries = atol(q);
				free(r);
				free(q);
				if (secsago > 0) {
					slow_queries_delta = (slow_queries < last_slow_queries) ? slow_queries : slow_queries - last_slow_queries;
					slow_queries_per_sec = 1.0 * slow_queries_delta / secsago;
				}
				if (absolute == 0) {
					status = get_status(slow_queries_per_sec, my_thresholds);
				} else {
					status = get_status((double)slow_queries_delta, my_thresholds);
				}
				pluginexit(status, "%ld slow queries in %ld seconds (%.4f/sec)",
				    slow_queries_delta, secsago, slow_queries_per_sec);
			}
			perfdata("slow_queries_rate=%.4f;%s;%s slow_queries=%ld;%s;%s", slow_queries_per_sec, userwarning, usercritical, slow_queries_delta, userwarning, usercritical);
                } else if (strcmp(mode, "bufferpool-wait-free") == 0) {
                        char *r, *q;
                        double bufferpool_wait_free_per_sec = 0.0;
                        long  bufferpool_wait_free_delta = 0.0;
                        long  bufferpool_wait_free = 0;
                        long  last_bufferpool_wait_free = 0; 
                        time_t secsago;

                        interval = (interval > 0) ? interval : 0;
                        force_thresholds(&my_thresholds, &userwarning, &usercritical, "1", "10");
                        q = mysql_status_historical(&mysql, "Innodb_buffer_pool_wait_free", &secsago, interval);
                        r = mysql_status(&mysql, "Innodb_buffer_pool_wait_free");
                        if (r == NULL) { 
                                pluginexit(STATE_CRITICAL, "No connection information available.");
                        } else if (q == NULL) {
                                bufferpool_wait_free = atol(r);
                                free(r);
                                pluginexit(STATE_UNKNOWN, "no historical data");
                        } else {
                                bufferpool_wait_free = atol(r);
                                last_bufferpool_wait_free = atol(q);
                                free(r);
                                free(q);
                                if (secsago > 0) {
					bufferpool_wait_free_delta = (bufferpool_wait_free < last_bufferpool_wait_free) ? bufferpool_wait_free : bufferpool_wait_free - last_bufferpool_wait_free;
                                        bufferpool_wait_free_per_sec = 1.0 * bufferpool_wait_free_delta / secsago;
                                }
				if (absolute == 0) {
                                	status = get_status(bufferpool_wait_free_per_sec, my_thresholds);
				} else {
                                	status = get_status((double)bufferpool_wait_free_delta, my_thresholds);
				}
                                pluginexit(status, "%ld Innodb buffer pool waits in %ld seconds (%.4f/sec)",
				    bufferpool_wait_free_delta, secsago, bufferpool_wait_free_per_sec);
                        }
                        perfdata("bufferpool_free_waits_rate=%.4f;%s;%s bufferpool_free_waits=%ld;%s;%s", bufferpool_wait_free_per_sec, userwarning, usercritical, bufferpool_wait_free_delta, userwarning, usercritical);
                } else if (strcmp(mode, "log-waits") == 0) {
                        char *r, *q;
                        double log_waits_per_sec = 0.0;
                        long  log_waits_delta = 0;
                        long  log_waits = 0;
                        long  last_log_waits = 0;
                        time_t secsago;

                        interval = (interval > 0) ? interval : 0;
                        force_thresholds(&my_thresholds, &userwarning, &usercritical, "1", "10");
                        q = mysql_status_historical(&mysql, "Innodb_log_waits", &secsago, interval);
                        r = mysql_status(&mysql, "Innodb_log_waits");
                        if (r == NULL) {
                                pluginexit(STATE_CRITICAL, "No connection information available.");
                        } else if (q == NULL) {
                                log_waits = atol(r);
                                free(r);
                                pluginexit(STATE_UNKNOWN, "no historical data");
                        } else {
                                log_waits = atol(r);
                                last_log_waits = atol(q);
                                free(r);
                                free(q);
                                if (secsago > 0) {
                                        log_waits_delta = (log_waits < last_log_waits) ? log_waits : log_waits - last_log_waits;
                                        log_waits_per_sec = 1.0 * log_waits_delta / secsago;
                                }
				if (absolute == 0) {
                                	status = get_status(log_waits_per_sec, my_thresholds);
				} else {
                                	status = get_status((double)log_waits_delta, my_thresholds);
				}
                                pluginexit(status, "%ld Innodb log write requests waiting in %ld seconds (%.4f/sec)",
                                    log_waits_delta, secsago, log_waits_per_sec);
                        }
			perfdata("log_waits_rate=%.4f;%s;%s log_waits=%ld;%s;%s", log_waits_per_sec, userwarning, usercritical, log_waits_delta, userwarning, usercritical);
                } else if (strcmp(mode, "qcache-lowmem-prunes") == 0) {
                        char *r, *q;
                        double lowmem_prunes_per_sec = 0.0;
                        long  lowmem_prunes_delta = 0;
                        long  lowmem_prunes = 0;
                        long  last_lowmem_prunes = 0;
                        time_t secsago;

                        interval = (interval > 0) ? interval : 0;
                        force_thresholds(&my_thresholds, &userwarning, &usercritical, "1", "10");
                        q = mysql_status_historical(&mysql, "Qcache_lowmem_prunes", &secsago, interval);
                        r = mysql_status(&mysql, "Qcache_lowmem_prunes");
                        if (r == NULL) {
                                pluginexit(STATE_CRITICAL, "No connection information available.");
                        } else if (q == NULL) {
                                lowmem_prunes = atol(r);
                                free(r);
                                pluginexit(STATE_UNKNOWN, "no historical data");
                        } else {
                                lowmem_prunes = atol(r);
                                last_lowmem_prunes = atol(q);
                                free(r);
                                free(q);
                                if (secsago > 0) {
                                        lowmem_prunes_delta = (lowmem_prunes < last_lowmem_prunes) ? lowmem_prunes : lowmem_prunes - last_lowmem_prunes;
                                        lowmem_prunes_per_sec = 1.0 * lowmem_prunes_delta / secsago;
                                }
				if (absolute == 0) {
                                	status = get_status(lowmem_prunes_per_sec, my_thresholds);
				} else {
                                	status = get_status((double)lowmem_prunes_delta, my_thresholds);
				}
                                pluginexit(status, "%ld Qcache low memory prunes in %ld seconds (%.4f/sec)",
                                    lowmem_prunes_delta, secsago, lowmem_prunes_per_sec);
                        }
                        perfdata("lowmem_prunes_rate=%.4f;%s;%s lowmem_prunes=%ld;%s;%s", lowmem_prunes_per_sec, userwarning, usercritical, lowmem_prunes_delta, userwarning, usercritical);
		} else {
			pluginexit(STATE_UNKNOWN, "unknown mode %s", mode);
		}

		/* close the connection */
		mysql_close (&mysql);

	}
	printf(_("%s - %s|%s"), severities[exitcode], exitmessagestr, perfdatastr);
	printf("\n");
	return exitcode;
}


/* process command-line arguments */
int
process_arguments (int argc, char **argv)
{
	int c;
	char *warning = NULL;
	char *critical = NULL;

	int option = 0;
	static struct option longopts[] = {
		{"hostname", required_argument, 0, 'H'},
		{"database", required_argument, 0, 'd'},
		{"username", required_argument, 0, 'u'},
		{"password", required_argument, 0, 'p'},
		{"port", required_argument, 0, 'P'},
		{"verbose", no_argument, 0, 'v'},
		{"version", no_argument, 0, 'V'},
		{"help", no_argument, 0, 'h'},
		{"mode", 1, 0, 'm'},
		{"warning", required_argument, 0, 'w'},
		{"critical", required_argument, 0, 'c'},
		{"interval", required_argument, 0, 'i'},
		{"absolute", no_argument, 0, 'a'},
		{0, 0, 0, 0}
	};

	if (argc < 1)
		return ERROR;

	while (1) {
		c = getopt_long (argc, argv, "hvVSP:p:u:d:H:m:w:c:i:a", longopts, &option);

		if (c == -1 || c == EOF)
			break;

		switch (c) {
		case 'H':									/* hostname */
			if (is_host (optarg)) {
				db_host = optarg;
			}
			else {
				usage2 (_("Invalid hostname/address"), optarg);
			}
			break;
		case 'd':									/* hostname */
			db = optarg;
			break;
		case 'u':									/* username */
			db_user = optarg;
			break;
		case 'p':									/* authentication information: password */
			asprintf(&db_pass, "%s", optarg);

			/* Delete the password from process list */
			while (*optarg != '\0') {
				*optarg = 'X';
				optarg++;
			}
			break;
		case 'P':									/* critical time threshold */
			db_port = atoi (optarg);
			break;
		case 'v':
			verbose++;
			break;
		case 'V':									/* version */
			/*
			print_my_revision (progname, revision);
			*/
			printf (_(MY_REVISION), progname);
			exit (STATE_OK);
		case 'h':									/* help */
			print_help ();
			exit (STATE_OK);
		case 'm':
			asprintf(&mode, "%s", optarg);
			break;
		case 'w':
			userwarning = optarg;
			break;
		case 'c':
			usercritical = optarg;
			break;
		case 'i':									/* interval to watch values growing */
			interval = atol (optarg);
			break;
		case 'a':									/* output difference instead of rate */
			absolute = 1;
			break;
		case '?':									/* help */
			usage5 ();
		}
	}

	c = optind;
	return validate_arguments ();
}


int
validate_arguments (void)
{
	if (mode == NULL)
		usage("Must specify a mode");

	if (db_user == NULL)
		db_user = strdup("");

	if (db_host == NULL)
		db_host = strdup("");

	if (db_pass == NULL)
		db_pass == strdup("");

	if (db == NULL)
		db = strdup("");

	return OK;
}


void
print_help (void)
{
	char *myport;
	asprintf (&myport, "%d", MYSQL_PORT);
/*
	print_revision (progname, revision);

	printf (_(COPYRIGHT), copyright, email);
*/
	printf (_(MY_REVISION), progname);

	printf (_(MY_COPYRIGHT), copyright, email);

	printf ("%s\n", _("This program checks various database parameters against threshold levels"));

  printf ("\n\n");

	print_usage ();

	printf (_(UT_HELP_VRSN));
	printf (" -m, --mode=STRING\n");
	printf ("    %s\n", _("where STRING is one of:"));
	printf ("    %s\n", _("slave-lag             (Check 'Seconds behind master')"));
	printf ("    %s\n", _("slave-io-running      (Check for 'Slave io running: Yes')"));
	printf ("    %s\n", _("slave-sql-running     (Check for 'Slave sql running: Yes')"));
	printf ("    %s\n", _("threads-connected     (Check 'Threads connected')"));
	printf ("    %s\n", _("threadcache-hitrate   (Check 'Thread cache hitrate')"));
	printf ("    %s\n", _("querycache-hitrate    (Check 'Query cache hitrate')"));
	printf ("    %s\n", _("keycache-hitrate      (Check 'MyISAM Key cache hitrate')"));
	printf ("    %s\n", _("bufferpool-hitrate    (Check 'InnoDB Buffer pool hitrate')"));
	printf ("    %s\n", _("tablecache-hitrate    (Check 'Table cache hitrate')"));
	printf ("    %s\n", _("table-lock-contention (Check 'Table lock contention')"));
	printf ("    %s\n", _("temp-disk-tables      (Check 'Percent of temp tables created on disk')"));
	printf ("    %s\n", _("connection-time       (Check 'Seconds to get a connection')"));
	printf ("    %s\n", _("slow-queries          (Check 'Slow queries'*)"));
	printf ("    %s\n", _("qcache-lowmem-prunes  (Check 'Query cache entries pruned because of low memory'*)"));
	printf ("    %s\n", _("bufferpool-wait-free  (Check 'InnoDB Buffer pool waits for clean page available'*)"));
	printf ("    %s\n", _("log-waits             (Check 'InnoDB log waits because of a too small log buffer'*)"));
	printf (_(UT_WARN_CRIT_RANGE));
	printf (_(UT_HOST_PORT), 'P', myport);
	printf (" -i, --interval=INTEGER\n");
	printf ("    %s\n", _("The window (in seconds) where rising values are measured"));
	printf (" -a, --absolute\n");
	printf ("    %s\n", _("Output the absolute difference of the two measurements instead of"));
	printf ("    %s\n", _("the change rate where the absolute difference is divided by time"));
	printf (" -d, --database=STRING\n");
	printf ("    %s\n", _("Database to check"));
	printf (" -u, --username=STRING\n");
	printf ("    %s\n", _("Username to login with"));
	printf (" -p, --password=STRING\n");
	printf ("    %s\n", _("Password to login with"));
	printf ("    ==> %s <==\n", _("IMPORTANT: THIS FORM OF AUTHENTICATION IS NOT SECURE!!!"));

	printf ("\n");

	printf ("%s\n", _("For extra security, create a user with minimal access."));
	printf ("Temporary files will be saved in %s\n", STR(TEMP_DIR));

/*
	printf (_(UT_SUPPORT));
*/
	printf (_(MY_UT_SUPPORT));
}


void
print_usage (void)
{
  printf (_("Usage:"));
	printf ("%s -m mode [-w warn] [-c crit]\n",progname);
  printf ("[-d database] [-H host] [-P port] [-u user] [-p password]\n");
}


void
force_thresholds (thresholds **thresholds, char **warning, char **critical,
    const char *defwarning, const char *defcritical) {
	if (*warning == NULL)
		*warning = defwarning;
	if (*critical == NULL)
		*critical = defcritical;
	set_thresholds(&my_thresholds, *warning, *critical);
}


char
*run_query (MYSQL *m, char *cmd, int col)
{
	MYSQL_ROW  row;
	MYSQL_RES *res;
	char *result = NULL;
  
	prdebug("Field %d of query %s\n", col, cmd);
	if (mysql_real_query(m, cmd, strlen(cmd)) != 0) {
		prdebug("CRIT: Query %s failed: %s\n", cmd, mysql_error(m));
	} else {
		res = mysql_store_result(m);
		if (!res) {
			prdebug("CRIT: Query %s failed: %s\n", cmd, mysql_error(m));
		} else if (row = mysql_fetch_row(res)) {
			unsigned int   num_fields = mysql_num_fields(res);
			unsigned long *len        = mysql_fetch_lengths(res);
	   
			prdebug("NOTICE: we have results\n");
			mysql_free_result(res);
			if (col < num_fields) {
				prdebug("Result column %d returns value %s (length %d)\n",
			    	    col, row[col], len[col]);
				result = len[col] > 0 ? strndup(row[col], len[col]) : NULL;
			} else {
				prdebug("CRIT: Query cmd returns %d fields (%d expected).\n",
			    	    num_fields, col);
			}
		}
	}
	return result;
}


char
*mysql_status (MYSQL *m, char *variable)
{
	char *result = NULL;
	char *cmd;
  
	prdebug("Status variable %s\n", variable);
	asprintf(&cmd, "SHOW /*!50000 global */ STATUS LIKE '%s'", variable);
        result = run_query(m, cmd, 1);
	free(cmd);
	if ((interval == 0) && (result != NULL)) {
		save_variable (variable, result);
	}
	return result;
}


char
*mysql_variable (MYSQL *m, char *variable)
{
	char *result = NULL;
	char *cmd;
  
	prdebug("Variable %s\n", variable);
	asprintf(&cmd, "SHOW /*!50000 global */ VARIABLES LIKE '%s'", variable);
        result = run_query(m, cmd, 1);
	free(cmd);
	if ((interval == 0) && (result != NULL)) {
		save_variable (variable, result);
	}
	return result;
}


char
*mysql_status_historical (MYSQL *m, char *variable, long *age, int desired_age)
{
	char *historical_value = NULL;
	if (desired_age == 0) {
		prdebug("want saved value of %s\n", variable);
		return load_variable(variable, age);
	} else {
		historical_value = mysql_status(m, variable);
		prdebug("sleeping %ld seconds\n", desired_age);
		sleep(desired_age);
		*age = desired_age;
		return historical_value;
	}
}


void
save_variable (char *variable, char  *value)
{
	FILE *fp;
	char *tmpfile = NULL;
	struct timeval now;
	
	asprintf(&tmpfile, "%s/%s_%s",
	    STR(TEMP_DIR), runident, variable);
	if (fp = fopen(tmpfile, "w")) {
		gettimeofday(&now, 0);
		fprintf(fp, "%ld %s %s\n", now.tv_sec, variable, value);
		fclose(fp);
		prdebug("Saved variable %s (%s) to %s\n", variable, value, tmpfile);
	} else {
		prdebug("Could not write to temp file %s\n", tmpfile);
	}
}


char 
*load_variable (char *variable, long *age)
{
	FILE *fp;
	char *tmpfile = NULL;
	struct timeval now;
	time_t when = 0;
	char *savedvalue = NULL;
	
	asprintf(&tmpfile, "%s/%s_%s",
	    STR(TEMP_DIR), runident, variable);
	if (fp = fopen(tmpfile, "r")) {
		savedvalue = malloc(MAX_INPUT_BUFFER);
		fscanf(fp, "%ld %*s %s", &when, savedvalue);
		fclose(fp);
		prdebug("Load variable %s (%s) from %s\n", variable, savedvalue, tmpfile);
	}
	gettimeofday(&now, 0);
	*age = now.tv_sec - when;
	return savedvalue;
}


char
*warning_range(thresholds *my_threshold)
{
	char *range = NULL;
	if (! my_threshold) {
		return NULL;
	} else if (! my_threshold->warning) {
		return NULL;
	} else {
		if (my_threshold->warning->start == 0) {
			asprintf(&range, "%s%lf",
				(my_threshold->warning->alert_on == INSIDE ? "@" : ""),
				my_threshold->warning->end);
		} else if (my_threshold->warning->end == 0) {
			asprintf(&range, "%s%lf:",
				(my_threshold->warning->alert_on == INSIDE ? "@" : ""),
				my_threshold->warning->start);
		} else if (my_threshold->warning->end != 0) {
			asprintf(&range, "%s%.lf:%.lf",
				(my_threshold->warning->alert_on == INSIDE ? "@" : ""),
				my_threshold->warning->start,
				my_threshold->warning->end);
		} else {
			asprintf(&range, "%s%lf",
				(my_threshold->warning->alert_on == INSIDE ? "@" : ""),
				my_threshold->warning->start);
		}
	}
	return range;
}

